function finalModel = LASSO_stability_CV(data,labels,dataSetName,shuffle,shuffleId,style,ppp,modelRank,classifierType)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if nargin < 9
    classifierType = 'SLR';
    classifierType = 'SVM';
    classifierType = 'LDA';
end

disp(['Classifier: ' classifierType])

if nargin < 3 || nargin > 9 || nargin == 4
    error('Number of input arguments must be 3, 5, 6, 7, or 8!')
end
if nargin == 3
    shuffle = 'off';
    shuffleId = [];
end
if nargin < 6
    style = 'normal';
end

init_params_classification;
[DataCV,data,labels,outerLoopSize,subj_index] = divide_data_into_CV_blocks(data,labels,style,nFolds,LOSO,subj_index,saveMem);

if plotFigs
    figure;hold on;ylim([0.3 0.8]);grid on
end


%col = lines(outerLoopSize);
clear lineName
test_accuracy = NaN*zeros(size(1,outerLoopSize));
accuracy_max = NaN*zeros(size(1,outerLoopSize));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Outer CV-loop to determine test error:
for s = 1:outerLoopSize
    if verbose_on
        disp(['Training classifier for CV block ' num2str(s) '/' num2str(outerLoopSize) '...'])
    end
    
    [dat,lab,labels_test,data_test,datOrig] = pick_training_and_test_data(DataCV,data,style,saveMem,s);
    
    %    options = glmnetSet;
    options = [];
    options.nlambda = nLambda;
    options.nrClasses = nrClasses;
    options.stability = stability;
    if strcmp(shuffle,'on')
        disp('Shuffling labels...')
        rng(500+shuffleId)
        rp = randperm(length(lab));
        lab = lab(rp);
    end
    
    % divide training data further to validation and training data.
    % Also perform feature transformations to the training and test data if
    % needed.
    [DataCVTrain,dat,data_test,datO,UUfinal,VVfinal] = ... 
        pick_training_and_validation_data(dat,data_test,lab,datOrig,nFoldsTrain,style,saveMem,modelRank,nrICs);
    
    if strcmp(style,'bilinear') | strcmp(style,'temporalbilinear')
        iter = 1;
        spars_lev = [];
        lambda_pairs = [];
        labels_estim = [];
        Classifier = [];
        for mmm = 1:length(lambdas_comp)
            for nnn = 1:length(lambdas_freq)
                % train classifier:         
                dtrain = permute(dat{1},[2 3 1]);
                dtest = permute(data_test{1},[2 3 1]);
                Classifier{iter} = train_sparse_bilinear(dtrain,...
                    1+lab,lambdas_freq(nnn),lambdas_comp(mmm),modelRank);
                % validate classifier:
                ytest = 1 + labels_test;
                resu = test_sparse_bilinear(dtest,ytest,Classifier{iter});
                
                ytrain = 1 + lab;
                resu_tr = test_sparse_bilinear(dtrain,ytrain,Classifier{iter});
                
                spars_lev(iter,:) = [sum(Classifier{iter}.W{1}(2:end)==0)/length(Classifier{iter}.W{1}(2:end)) sum(Classifier{iter}.V{1}(2:end)==0)/length(Classifier{iter}.V{1}(2:end))];
                
                disp(['  Test accuracy for lambdas (' num2str(lambdas_freq(nnn)) ...
                    ',' num2str(lambdas_comp(mmm))  '): ' num2str(resu.acc) ...
                    ', training accuracy: ' num2str(Classifier{iter}.ACC) ...
                    ' (' num2str(resu_tr.acc) ')' ' sparsity ' num2str(spars_lev(iter,:))])
                
                % save validation accuracy and lambda values:
                labels_estim(:,iter) = resu.labelsEstim;
                lambda_pairs(iter,:) = [lambdas_freq(nnn) lambdas_comp(mmm)];
                iter = iter + 1;
            end
        end
        finalModel(s).fit = Classifier;
        labels_TEST = (~labels_test) + 1;
        labels_true = repmat(labels_TEST,1,size(labels_estim,2));
        acc_test = sum(labels_estim == labels_true,1)/size(labels_true,1);
        disp(num2str(acc_test))
        disp(num2str(spars_lev(:,1)'))
        
        finalModel(s).labels_estim = labels_estim;
        finalModel(s).labels_test = labels_test;
        if options.stability
            finalModel(s).stability = lasso_stability(dat,lab,options);
        else
            finalModel(s).stability = [];
        end
        
        %%%%%%%%%%%%%%%%%%%%
    else
        if nrClasses == 2
            % RETURN THE FINAL MODEL CANDIDATES FOR THE OUTER FOLD #s:
            switch classifierType
                case 'SLR'
                    mdl = 'binomial';
                    link = 'logit';
                    warning off
                    [B,finalModel(s).fit] = lassoglm(dat,lab,mdl,'NumLambda',nLambda);
                    warning on
                    finalModel(s).fit.beta = B;
                    B2 = [finalModel(s).fit.Intercept;B];
                    %            finalModel(s).fit ZZ= glmnet(dat,lab,'binomial',options);
                    labels_estim = NaN*zeros(size(data_test,1),size(B2,2));
                    for ii = 1:size(B2,2)
                        pred_estim = glmval(B2(:,ii),data_test,link);
                        labels_estim(:,ii) = pred_estim > 0.5;
                    end
                    options.lambda = finalModel(s).fit.Lambda;
                case 'SVM'
                    labels_estim = NaN*zeros(size(data_test,1),length(boxC));
                    for ii = 1:length(boxC)
                        if highDim % in case of high-dimensional data, use lambda regularized SVM:
                            finalModel(s).fit(ii).mdl = fitclinear(dat,lab,'ClassNames',[0,1],'Lambda',boxC(ii));
                        else
                            finalModel(s).fit(ii).mdl = fitcsvm(dat,lab,'Standardize',stand,'ClassNames',[0,1],'BoxConstraint',boxC(ii));
                        end
                        labels_estim(:,ii) = predict(finalModel(s).fit(ii).mdl,data_test);
                    end
                    options.boxC = boxC;
                case 'LDA'
                    labels_estim = NaN*zeros(size(data_test,1),length(deltas));
                    for ii = 1:length(deltas)
                        finalModel(s).fit(ii).mdl = fitcdiscr(dat,lab,'DiscrimType','diagLinear','Delta',deltas(ii),'SaveMemory','on','FillCoeffs','off');
%                        finalModel(s).fit(ii).mdl = fitcdiscr(dat,lab,'DiscrimType','linear','Gamma','Delta',deltas(ii),'SaveMemory','on','FillCoeffs','off');
                        labels_estim(:,ii) = predict(finalModel(s).fit(ii).mdl,data_test);
                    end
                    
                    options.gammas = gammas;
                    options.deltas = deltas;
                    options.gammas_deltas = gammas_deltas;
            end            
        else            
            mdl = 'multinomial';
            link = 'logit';
            lab2 = zeros(length(lab),nrClasses);
            for e = 1:nrClasses
                lab2(find(lab==e),e) = 1;
            end            
            warning off
            finalModel(s).fit = glmnet(dat,lab2,mdl,options);
            warning on
            labels_estim = glmnetPredict(finalModel(s).fit,data_test,[],'class');
            labels_estim = reshape(labels_estim,size(data_test,1),size(finalModel(s).fit.a0,2));            
            options.lambda = finalModel(s).fit.lambda;
        end
        labels_true = repmat(labels_test,1,size(labels_estim,2));
        acc_test = sum(labels_estim == labels_true,1)/size(labels_true,1);
        finalModel(s).labels_estim = labels_estim;
        finalModel(s).labels_test = labels_test;
        if options.stability
            finalModel(s).stability = lasso_stability(dat,lab,options);
        else
            finalModel(s).stability = [];
        end
        
    end
    clear Models
    Acc = NaN*zeros(nFoldsTrain,nLambda);
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%5
    % HYPERPARAMETER OPTIMIZATION LOOP:
    for nn = 1:nFoldsTrain
        
        [data_tr,labels_tr,data_te,labels_te,UU,VV] = pick_innerfold_data(DataCVTrain,style,saveMem,datO,nn,nrICs,modelRank);
        
        if strcmp(style,'bilinear') | strcmp(style,'temporalbilinear')
            iter = 1;
            lambda_pairs = [];
            labels_estim = [];
            Classifier = [];
            for mmm = 1:length(lambdas_comp)
                for nnn = 1:length(lambdas_freq)
                    dd = permute(data_tr{1},[2 3 1]);
                    dt = permute(data_te{1},[2 3 1]);
                    yte = 1 + labels_te;
                    ytr = 1 + labels_tr;
                    % train classifier:
                    Classifier{iter} = train_sparse_bilinear(dd,...
                        1+labels_tr,lambdas_freq(nnn),lambdas_comp(mmm),modelRank);
                    % validate classifier:
                    resu_tr = test_sparse_bilinear(dd,ytr,Classifier{iter});
                    resu = test_sparse_bilinear(dt,yte,Classifier{iter});
                    %      disp(['  Test accuracy for lambdas (' num2str(lambdas_freq(nnn)) ',' num2str(lambdas_comp(mmm))  '): ' num2str(resu.acc) ', training accuracy: ' num2str(Classifier{iter}.ACC) ' (' num2str(resu_tr.acc) ')'])
                    %                    Classifier{iter}
                    % save validation accuracy and lambda values:
                    labels_estim(:,iter) = resu.labelsEstim;
                    lambda_pairs(iter,:) = [lambdas_freq(nnn) lambdas_comp(mmm)];
                    iter = iter + 1;
                end
            end
            Models(nn).fits = Classifier;
            labels_TEST = (~labels_te) + 1;
            labels_true = repmat(labels_TEST,1,size(labels_estim,2));
            acc = sum(labels_estim == labels_true,1)/size(labels_true,1);
            Acc(nn,1:length(acc)) = acc;
        else
            if nrClasses == 2
                switch classifierType
                    case 'SLR'
                        mdl = 'binomial';
                        link = 'logit';
                        warning off
                        [B,Models(nn).fits] = lassoglm(data_tr,labels_tr,mdl,'NumLambda',nLambda);
                        warning on
                        Models(nn).fits.beta = B;
                        B2 = [Models(nn).fits.Intercept;B];
                        %    Models(nn).fits = glmnet(data_tr, labels_tr, 'binomial', options);
                        labels_estim = NaN*zeros(size(data_te,1),size(B2,2));
                        for ii = 1:size(B2,2)
                            pred_estim = glmval(B2(:,ii),data_te,link);
                            labels_estim(:,ii) = pred_estim > 0.5;
                        end
                    case 'SVM'
                        labels_estim = NaN*zeros(size(data_te,1),length(boxC));
                        for ii = 1:length(boxC)
                            if highDim % for high-dimensional data, use lambda-regularized SVM:
                                Models(nn).fit(ii).mdl = fitclinear(data_tr,labels_tr,'ClassNames',[0,1],'Lambda',boxC(ii));
                            else
                                Models(nn).fit(ii).mdl = fitcsvm(data_tr,labels_tr,...
                                    'Standardize',stand,'ClassNames',[0,1],'BoxConstraint',boxC(ii));
                            end
                            labels_estim(:,ii) = predict(Models(nn).fit(ii).mdl,data_te);
                        end
                    case 'LDA'
                        
                        labels_estim = NaN*zeros(size(data_te,1),length(deltas));
                        for ii = 1:length(deltas)
                            Models(nn).fit(ii).mdl = fitcdiscr(data_tr,labels_tr,'DiscrimType','diaglinear','Delta',deltas(ii),'SaveMemory','on','FillCoeffs','off');
                            labels_estim(:,ii) = predict(Models(nn).fit(ii).mdl,data_te);
                        end
                        
                        %labels_estim = NaN*zeros(size(data_te,1),length(gammas_deltas));
                        %for ii = 1:length(gammas_deltas)
                        %    if highDim % for high-dimensional data, set LDA's savememory option 'on':
                        %        Models(nn).fit(ii).mdl = fitcdiscr(data_tr,labels_tr,'Gamma',gammas_deltas(ii,1),'Delta',gammas_deltas(ii,2),'SaveMemory','on','FillCoeffs','off');
                        %    else
                        %        Models(nn).fit(ii).mdl = fitcdiscr(data_tr,labels_tr,'Gamma',gammas_deltas(ii,1),'Delta',gammas_deltas(ii,2));
                        %    end
                        %    labels_estim(:,ii) = predict(Models(nn).fit(ii).mdl,data_te);
                        %end
                        options.gammas_deltas = gammas_deltas;
                        options.gammas = gammas;
                        options.deltas = deltas;
                        
                end
            else                
                labels_tr2 = zeros(length(labels_tr),nrClasses);
                for e = 1:nrClasses
                    labels_tr2(find(labels_tr==e),e) = 1;
                end
                warning on
                Models(nn).fits = glmnet(data_tr, labels_tr2, 'multinomial', options);
                warning off
                labels_estim = glmnetPredict(Models(nn).fits,data_te,[],'class');
                labels_estim = reshape(labels_estim,size(data_te,1),size(Models(nn).fits.a0,2));
            end
            labels_true = repmat(labels_te,1,size(labels_estim,2));
            acc = sum(labels_estim == labels_true,1)/size(labels_true,1);
            Acc(nn,1:length(acc)) = acc;
        end
    end
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % HYPERPARAMETER OPTIMIZATION LOOP ENDS
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % pick best model and corresponding test accuracy:
    accMean = mean(Acc,1);
    if length(accMean) > length(acc_test)
        accMean = accMean(1:length(acc_test));
    end
    [accBest,mai] = max(accMean);
    finalModel(s).bestInd = mai;
    finalModel(s).AccCV_all = Acc;
    finalModel(s).AccCV_mean = accMean;
    finalModel(s).AccCV_best = accBest;
    finalModel(s).acc_test_all = acc_test;
    finalModel(s).acc_test_best = max(acc_test);
    finalModel(s).acc_test = acc_test(mai);
    
    finalModel(s).ModelsCV = Models;
    test_accuracy(1,s) = acc_test(mai);
    accuracy_max(1,s) = max(acc_test);
    if verbose_on
        disp(['Test result CV: ' num2str(acc_test(mai))])
        disp(['Test result best: ' num2str(max(acc_test))])
    end
    switch style
        case {'temporalthreewayDA','threewayDA','temporalthreewaySDA',...
                'temporalthreewayRDA','temporalthreewayDAfull','temporalthreewayDAfullorig'}
            finalModel(s).UUfinal = UUfinal;
            finalModel(s).VVfinal = VVfinal;
            finalModel(s).modelRank = modelRank;
        case {'spectralLDA','temporalLDA','temporalRLDA','temporalLDAfull',...
                'temporalICA','temporalPCA','spectralICA','spatiotemporalLDA',...
                'spatiotemporalRLDA','spatiotemporalLDAfull',...
                'spatialLDA','spatialRLDA','spatialLDAfull'}
            finalModel(s).VVfinal = VVfinal;
            if strcmp(style,'spatiotemporalLDA') | strcmp(style,'spatiotemporalRLDA') | strcmp(style,'spatiotemporalLDAfull')
                finalModel(s).UUfinal = UUfinal;
            end
    end
end
