function [cp_struct,colBar] = colorplot_universal_modif(data,pos_map,P,scale,special_chnls,axes_h,cp_struct)

% todo: colorbar as output argument, nonlin color axis/bar
%
% P.scale = 0;  %0: [0 automax] 1:[automin automax] 2:[-max(abs) +max(abs)] 3:use parameter scale
% P.colormap = 'jet'; % define colormap (string or RGB-table), also 'BlueWhiteRed'
% P.nonlin = 0; %0: linear; 1: log10; 2: root; 3: square % not tested yet!
% P.res = 0.001; % def = 0.001 % resolution in m
% P.colorbar = 0; %0:don't show colorbar 1:show colorbar right 2:show colorbar down
% P.fast_update = 0; %0: normal plot, 1: for 2nd plot only data is updated (requires h)
% P.plot_dots = 3; % size of electrode grid dots, 0:no channle dots
% P.chnl_labels = 0; % 0:don't show channel labels, 1: show channel labels
% P.flat_plot = 1; %0: make plot flat, that electrode dots are visible 1:3D plot 2:cylinder 3d plot
% P.interpol = 'cubic'; %interpolation method: 'cubic', 'linear', 'nearest', 'v4'
% P.extrapol_val = 0; % value for extrapolation 0:0, 1: lower scale, 2: mean lower/higher scale
% P.border = 0.01;

%axes_h: optional axes handle where to plot (def.:current axes) (not needed for update)
%cp_struct: optional surf handle: needed for update

colBar = [];

P = bg_default(P,'scale',0,'colormap','jet','nonlin',0,'res',0.001,...
    'colorbar',0,'fast_update',0,'plot_dots',3,'chnl_labels',0,...
    'flat_plot',1,'interpol','cubic','extrapol_val',0,'border',0.01);

% define scaling of the plot:
switch P.scale
    case 0, scale = [0 max(data)];
    case 1, scale = [min(data) max(data)];
    case 2, scale = [-1 1] * max(abs(data));
    case 3, if size(scale) ~= [1 2], error('invalid scale');end
end

% nonlinear data rescale:
switch P.nonlin
    case 0, plot_scale = scale;
    case 1, data = log10(data/scale(2)); plot_scale = ...
            [log10(scale(1)/scale(2)),0]; % not tested
    case 2, data = sqrt(data); plot_scale = sqrt(scale); % not tested
    case 3, data = data.^2; plot_scale = scale.^2; % not tested
end

switch P.extrapol_val
    case 0, extrapol_val = 0;
    case 1, extrapol_val = scale(1);
    case 2, extrapol_val = mean(scale);
end

if ~P.fast_update
    if (nargin < 6), axes_h = gca;end
    % create meshgrid:
    %    P.border = 0.003;
    ti = min(pos_map(:,1))-P.border:P.res:max(pos_map(:,1))+P.border;
    tiy = min(pos_map(:,2))-P.border:P.res:max(pos_map(:,2))+P.border;
    if 0
        %        P.border = 0.00;
        ti = linspace(min(pos_map(:,1))-P.border,max(pos_map(:,1))+P.border,14);
        %       P.border = 0.00;
        tiy = linspace(min(pos_map(:,2))-P.border,max(pos_map(:,2))+P.border,4);
    end
    
    % Extend data for extrapolation:
    cp_struct.extend_x = [ti';ti';  ones(length(tiy)-2,1) * ti(1) ; ones(length(tiy)-2,1) * ti(end) ];
    cp_struct.extend_y = [ones(length(ti),1) * tiy(1) ; ones(length(ti),1) * tiy(end) ;tiy(2:end-1)' ;tiy(2:end-1)'];
    cp_struct.extend_val = extrapol_val * ones(length(cp_struct.extend_x),1); % set extrapolation points to lower scale; todo: better zero???
    
    [cp_struct.XI,cp_struct.YI] = meshgrid(ti,tiy);
end

% if size(data,1) > size(pos_map,1)
%   data = data(1:size(pos_map,1),:);
% end
% inter/extrapolate values in between defined values
CI = griddata([pos_map(:,1);cp_struct.extend_x],[pos_map(:,2);...
    cp_struct.extend_y],[data;cp_struct.extend_val],cp_struct.XI,...
    cp_struct.YI,P.interpol);

if P.fast_update % update existing plot
    if (nargin < 6) error('cp_struct needed for fast update');end
    set(cp_struct.surf_h,'CData',CI); % update colorplot
    caxis(get(cp_struct.surf_h,'parent'),plot_scale);
    %if ~P.flat_plot, set(surf_h,'ZData',ZI);end % update Z-axis
    
else % make a new plot
    if ~exist('special_chnls'), special_chnls = [];end
    
    if isstr(P.colormap) & strcmp(lower(P.colormap),'bluewhitered')
        P.colormap = [linspace(0,1,32)',linspace(0,1,32)',...
            linspace(1,1,32)';linspace(1,1,32)',linspace(1,0,32)',linspace(1,0,32)'];
    end
    
    if (P.flat_plot == 2)           % cylinder plot:
        circ = (ti(end) - ti(1)); % circumface
        r = circ / (2*pi); % radius
        h = tiy(end) - tiy(1); % high
        
        pos_map_3d(:,3) = pos_map(:,2);
        pos_map_3d(:,1) = r * cos( (pos_map(:,1)-ti(1)) * 2 * pi / circ  );
        pos_map_3d(:,2) = r * sin( (pos_map(:,1)-ti(1)) * 2 * pi / circ  );
        
        R = r * ones(1,size(CI,1)); % for cylinder all equal
        
        [cp_struct.XI,cp_struct.YI,cp_struct.ZI] = cylinder(R,size(CI,2)); % create cylnder with unit Z-dimension [0 1]
        
        cp_struct.ZI = cp_struct.ZI * h + tiy(1); % adjust Z- scaling and position
        
    else                            %flat colorplot
        cp_struct.ZI = zeros(size(CI));
    end
    
    hold(axes_h, 'off');
    colormap(P.colormap);
    cp_struct.surf_h = surf(axes_h,cp_struct.XI,cp_struct.YI,...
        cp_struct.ZI,CI,'LineStyle','none'); % SURF-PLOT ******
    
    if strcmp(P.colormap,'hsv') && length(data) == 64
        data2 = [0; data];
        data2 = reshape(data2,13,5)';
        data2 = flipud(data2);
        data2(2:2:end,:) = fliplr(data2(2:2:end,:));
        %    data2(2:2:end,:) = [data2([2:2:end],2:end) data2([2:2:end],1)];
        hold(axes_h, 'off');
        colormap(P.colormap);
        cp_struct.surf_h = imagesc(data2); % SURF-PLOT ******
    end
    if 0
        %data2 = [0; data];
        data2 = reshape(data(1:48),12,4)';
        data2 = flipud(data2);
        data2(2:2:end,:) = fliplr(data2(2:2:end,:));
        %    data2(2:2:end,:) = [data2([2:2:end],2:end) data2([2:2:end],1)];
        hold(axes_h, 'off');
        colormap(P.colormap);
        cp_struct.surf_h = imagesc(data2); % SURF-PLOT ******
    end
    
    plot_scale = sort(plot_scale);
    caxis(axes_h, plot_scale);
    if P.flat_plot == 1
        view(axes_h, 2);
        axis(axes_h, 'off');
    else
        alpha(axes_h, .7);
    end% whatch from top
    axis(axes_h, 'equal');
    axis(axes_h, 'tight');
    %     if P.flat_plot == 1
    %         set(surf_h,'ZData',zeros(size(ZI)));% make plot flat to make Electrode dots visible
    %         set(surf_h,'CData',ZI); % needed, because preveous line seems to change CData as well
    %     end
    
    % plot colorbar:
    switch P.colorbar
        case 0, % do nothing
        case 1, colBar = colorbar('peer',axes_h,'location','EastOutside');
        case 2, colBar = colorbar('peer',axes_h,'location','SouthOutside');
    end
    
    % if (P.colorbar == 1), colorbar('location','East');end
    
    if P.plot_dots % plot channel dots
        hold(axes_h, 'on');
        if P.flat_plot == 2;
            plot3(axes_h, pos_map_3d(:,1),pos_map_3d(:,2),...
                pos_map_3d(:,3),'o','MarkerEdgeColor','k','MarkerFaceColor','k','MarkerSize',P.plot_dots);
        else
            plot(axes_h, pos_map(:,1),pos_map(:,2),'o',...
                'MarkerEdgeColor','k','MarkerFaceColor','k','MarkerSize',P.plot_dots);
        end
        if ~isempty(special_chnls) % make some channel dots red
            plot(axes_h, pos_map(special_chnls,1),...
                pos_map(special_chnls,2),'o','MarkerEdgeColor','k','MarkerFaceColor','r','MarkerSize',3);
        end
    end
    
    if P.chnl_labels % label channel dots
        for i =1:size(pos_map,1)
            text(pos_map(i,1)+0.02,pos_map(i,2)-0.03,[PREF.ELECTRODE_CONF.names{i}],'FontSize',10,'BackgroundColor','w', 'parent', axes_h);
        end
    end
end

function O = bg_default(P,varargin)

O = [];

if iscell(P)
    if length(P) == 1
        % cell contains one struct
        O = P{1};
    else
        % interpret P as cell with the form 'field1',arg1,'field2',arg2,....
        k = 1;
        while k < length(P)
            O = setfield(O,P{k},P{k+1});
            k = k+2;
        end
    end
    
else
    O = P;
end

k = 1;
while k < length(varargin)    
    field = varargin{k};
    arg = varargin{k+1};
    if ~isfield(O,field), O = setfield(O,field,arg); end    
    k = k+2;    
end


