function features = compute_projections(f,data)
% This function extracts Spectral LDA features from data using linear projections.
% 
%
% Inputs:
%
% f - coefficient vectors, i.e., the directions of projections.
% data - a data set to be projected. data is a C*N*F matrix, where:
% C = the number of ICs or channels
% N = the number of time-windows
% F = the number of frequency bins
%
% Outputs:
%
% features - projected epochs. These projections will be used as 
% classifier inputs

data = permute(data,[3 2 1]);

p = size(data,1);
n = size(data,2);
c = size(data,3);

nrFeatures = size(f,1); % number of estimated directions
  
dataProj = zeros(nrFeatures,c,n);
iter = 1;
for m = 1:nrFeatures 
  for k = 1:c
    w = squeeze(f(m,:,k));
    if sum(w) == 0
        w = randn(size(w));
    end
    w = w/norm(w);
    w = w(:);
    X = data(:,:,k)';
    Y = X*w;
    dataProj(m,k,:) = Y;
  end
end

features = tensor_to_matrix(dataProj);
