function [locations,locationsPhase] = get_locations(spatComp,Th,colorinds,nrChannels)

% Inputs:
%
% spatComp - M*N matrix, where M is the number of spatial patterns to be
% plotted in the same helmet/array, N is the number of sensors.
%
% Th - Scalar between 0 and 1 defining the threshold for spatial maps.
% Try different thresholds for visualization!


locationsPhase = [];

if nargin < 4
    nrChannels = [];
end
if nargin < 3
    colorinds = [];
end
if nargin < 2
    Th = [];
end
% compute magnitudes of the independent components:
if nrChannels == -1
    spatCompLoc = abs(spatComp);
else
    if ~isempty(nrChannels)
        nrChannelsHalf = floor(nrChannels/2);
        spatCompLoc = zeros(size(spatComp,1),nrChannelsHalf);
        for mm = 1:size(spatComp,1)
            spatCompLoc(mm,:) = sqrt(abs(spatComp(mm,1:nrChannelsHalf)).^2+abs(spatComp(mm,(nrChannelsHalf+1):nrChannels)).^2);
        end
    else
        if sum(abs(imag(spatComp))) ~= 0
            spatCompLoc = abs(spatComp);
            locations2 = angle(spatComp);
            %scale these to -pi/2 ... pi/2 because sign is indetermined anyway
            locationsPhase=locations2-pi*(abs(locations2)>pi/2).*sign(locations2);
        else
            spatCompLoc = abs(spatComp);
        end
    end
end
% check highest
if ~isempty(Th)
    [ma locations] = max(spatCompLoc,[],1);
    locations(ma <= Th) = 0;
else
    locations = spatCompLoc;
    
end
if ~isempty(colorinds)
    locations = -1*locations;
    un = nonzeros(fliplr(unique(locations)))';
    for m = 1:length(un)
        locations(locations == un(m)) = colorinds(m);
    end
end




