function [u,v,Uspher,Vspher] = regularized_threewayDA(X,labels,R,style)

nrIte = 10;
plotFig = 0;

if nargin == 0
    N = 100;
    R = 5;
    plotFig = 0;
    labels = [zeros(1,N/2) ones(1,N/2)];
    X1 = randn(8,5,N/2);
    X2 = 0.5 + randn(8,5,N/2)*0.8;
    X = cat(3,X1,X2);
end

%[w,J,Sb,Sw] = compute_LDA(X,labels,'diag');
%[w2,J2] = compute_LDA(X,labels,'none');

X1 = X(:,:,labels == 0);
X2 = X(:,:,labels == 1);

% INITIALIZATION:
E1 = mean(X1,3);
E2 = mean(X2,3);

if strcmp(style,'fullorig') % use original initialization procedure for 2D-LDA
    U = eye(size(E1,1),R);
    V = eye(size(E1,2),R);
    Vs{1} = V;
    Us{1} = U;
    v = V;
else % use mean difference initialization procedure for 2D-LDA
    [U,S,V] = svd(E2-E1);    
    Vs{1} = V(:,1:R);
    Us{1} = U(:,1:R);
    v = V(:,1:R);
end

for ite = 1:nrIte
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % COMPUTATIONS FOR FIXED V:
    
    % between-class scatter matrix:
    Sb = (E1-E2)*(v*v')*(E1-E2)';
    
    % within-class scatter matrix:
    N1 = size(X1,3);
    S1 = zeros(size(E1,1),size(E1,1));
    for n = 1:N1
        Z = X1(:,:,n);
        S1 = S1 + (Z-E1)*(v*v')*(Z-E1)';
    end
    N2 = size(X2,3);
    S2 = zeros(size(E2,1),size(E2,1));
    for n = 1:N2
        Z = X2(:,:,n);
        S2 = S2 + (Z-E2)*(v*v')*(Z-E2)';
    end
    Sw = S1 + S2;
    if strcmp(style,'diagonal')
        % use diagonal covariance:
        Sw = diag(diag(Sw));
        SwInv = diag(1./diag(Sw));
    elseif strcmp(style,'full')
        % use very slight regularization:
        tol = 1e-2;
        Sw = Sw + eye(size(Sw))*mean(diag(Sw))*tol;
        SwInv = inv(Sw);
    elseif strcmp(style,'fullorig')
        tol = 1e-2;
        Sw = Sw + eye(size(Sw))*mean(diag(Sw))*tol;
        SwInv = inv(Sw);
    else
        error('Wrong regularization "style"!')
    end
    % compute eigenvalues:
    C = SwInv*Sb;
    [u,e] = eigs(C,size(v,2));
    
    Us{ite+1} = u;
    if ~isreal(u)
       warning('Estimated factor "u" is not real!') 
    end
    noU(ite) = norm(abs(Us{ite+1}(:))-abs(Us{ite}(:)));
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % COMPUTATIONS FOR FIXED U:
    
    % between-class scatter matrix:
    Sb = (E1-E2)'*(u*u')*(E1-E2);
    
    % within-class scatter matrix:
    N1 = size(X1,3);
    S1 = zeros(size(E1,2),size(E1,2));
    for n = 1:N1
        Z = X1(:,:,n);
        S1 = S1 + (Z-E1)'*(u*u')*(Z-E1);
    end
    N2 = size(X2,3);
    S2 = zeros(size(E2,2),size(E2,2));
    for n = 1:N2
        Z = X2(:,:,n);
        S2 = S2 + (Z-E2)'*(u*u')*(Z-E2);
    end
    Sw = S1 + S2;
    if strcmp(style,'diagonal')
        % use diagonal covariance:
        Sw = diag(diag(Sw));
        SwInv = diag(1./diag(Sw));
    elseif strcmp(style,'full') | strcmp(style,'fullorig')
        % use very slight regularization:
        tol = 1e-2;
        Sw = Sw + eye(size(Sw))*mean(diag(Sw))*tol;
        SwInv = inv(Sw);
    else
        error('Wrong regularization "style"!')
    end
    % compute eigenvalues:
    C = SwInv*Sb;
    [v,e] = eigs(C,size(u,2));
    Vs{ite+1} = v;
    if ~isreal(v)
       warning('Estimated factor "v" is not real!') 
    end
    noV(ite) = norm(abs(Vs{ite+1}(:))-abs(Vs{ite}(:)));
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if plotFig
    % figure,plot(e)
    its = [1:7];
    figure;
    co = {'r--','g--','b--','c--','m--','k--','y--'};
    for fac = 1:R
        subplot(2,R,fac)
        hold on;
        for m = its
            title(['u (' num2str(fac) ')'])
            plot(abs(Us{m}(:,fac)),co{m});grid on
        end
    end
    for fac = 1:R
        subplot(2,R,fac+R)
        hold on;
        for m = its
            title(['v (' num2str(fac) ')'])
            plot(abs(Vs{m}(:,fac)),co{m});grid on
        end
    end
end

Uspher = Us{1};
Vspher = Vs{1};
