function [Classifier_final,resu] = train_and_test_classifier(data_train,data_validation,data_test,labels_train,labels_validation,labels_test,classifier_type,params)
% This function trains, validates and tests classifiers.
%
% Inputs:
%
% data_train - training data set
% data_validation - validation data set
% data_test - test data set
%
% data sets must be provided in "spectral format" with dimensions p*n*f, 
% where p is the number of components, n is the number of epochs, and f
% is the number frequency bins.
%
% labels_train - training data set class labels
% labels_validation - validation data set class labels
% labels_test - test data set class labels
% classifier_type - defines the classifier, and must be one of the
% following: 'sparse_bilinear', 'spectral_LDA', or 'three_way_DA'.
% params - a struct containg all necessary parameters of the classifiers.
%
% Outputs:
% 
% Classifier_final - parameters of the trained classifier
% resu - test results of the classifier


 
% **************************************************************
% Training and validation of classifiers:
% **************************************************************
switch classifier_type
    case 'sparse_bilinear'        
        disp('Training and testing Sparse bilinear classifier.')
        [resu,Classifier_final] = train_and_test_sparse_bilinear(data_train,data_validation,data_test,labels_train,labels_validation,labels_test,params.sparse_bilinear);
    case 'spectral_LDA'
        disp('Training and testing Spectral LDA classifier.')       
        if ~isfield(params,'spectral_LDA')
            params.spectral_LDA = [];
        end        
        [resu,Classifier_final] = train_and_test_spectral_LDA(data_train,data_validation,data_test,labels_train,labels_validation,labels_test,params.spectral_LDA);
    case 'three_way_DA'
        disp('Training and testing Three-way Discriminant analysis classifier.')
        [resu,Classifier_final] = train_and_test_three_way_DA(data_train,data_validation,data_test,labels_train,labels_validation,labels_test,params.three_way_DA);
    otherwise
        error('Unknown classifier type!')
end


